using System.Globalization;
using System.Runtime.CompilerServices;
using System.Xml.Linq;
using Microsoft.AspNetCore.Routing;
using Smartstore.Collections;
using Smartstore.Core.Configuration;
using Smartstore.Core.Data;
using Smartstore.Core.Identity;
using Smartstore.Core.Localization;
using Smartstore.Core.Stores;
using Smartstore.Data;
using Smartstore.Data.Hooks;
using Smartstore.IO;
using Smartstore.Threading;
using Smartstore.Utilities;

namespace Smartstore.Core.Seo
{
    public partial class XmlSitemapGenerator : AsyncDbSaveHook<BaseEntity>, IXmlSitemapGenerator
    {
        private const string SitemapsNamespace = "http://www.sitemaps.org/schemas/sitemap/0.9";
        private const string XhtmlNamespace = "http://www.w3.org/1999/xhtml";
        private const string SitemapFileNamePattern = "sitemap-{0}.xml";
        private const string SitemapIndexPathPattern = "sitemap.xml/{0}";
        private const string LockFileNamePattern = "sitemap-{0}-{1}.lock";

        /// <summary>
        /// The maximum number of sitemaps a sitemap index file can contain.
        /// </summary>
        private const int MaximumSiteMapCount = 50000;

        /// <summary>
        /// The maximum number of sitemap nodes allowed in a sitemap file. The absolute maximum allowed is 50,000 
        /// according to the specification. See http://www.sitemaps.org/protocol.html but the file size must also be 
        /// less than 10MB. After some experimentation, a maximum of 2.000 nodes keeps the file size below 10MB.
        /// </summary>
        internal const int MaximumSiteMapNodeCount = 2000;

        /// <summary>
        /// The maximum size of a sitemap file in bytes (10MB).
        /// </summary>
        private const int MaximumSiteMapSizeInBytes = 10485760;

        private readonly IEnumerable<Lazy<IXmlSitemapPublisher>> _publishers;
        private readonly SmartDbContext _db;
        private readonly ILanguageService _languageService;
        private readonly IUrlService _urlService;
        private readonly ICommonServices _services;
        private readonly ICustomerService _customerService;
        private readonly ILockFileManager _lockFileManager;
        private readonly LinkGenerator _linkGenerator;
        private readonly AsyncRunner _asyncRunner;

        private readonly IFileSystem _tenantRoot;
        private readonly string _baseDir;

        public XmlSitemapGenerator(
            IEnumerable<Lazy<IXmlSitemapPublisher>> publishers,
            SmartDbContext db,
            ILanguageService languageService,
            IUrlService urlService,
            ICommonServices services,
            ICustomerService customerService,
            ILockFileManager lockFileManager,
            LinkGenerator linkGenerator,
            AsyncRunner asyncRunner)
        {
            _publishers = publishers;
            _db = db;
            _languageService = languageService;
            _urlService = urlService;
            _services = services;
            _customerService = customerService;
            _lockFileManager = lockFileManager;
            _linkGenerator = linkGenerator;
            _asyncRunner = asyncRunner;

            _tenantRoot = _services.ApplicationContext.TenantRoot;
            _baseDir = "Sitemaps";
        }

        public ILogger Logger { get; set; } = NullLogger.Instance;

        #region Hook

        public override Task<HookResult> OnAfterSaveAsync(IHookedEntity entry, CancellationToken cancelToken)
        {
            if (entry.Entity is Store store && entry.InitialState != EntityState.Added)
            {
                Invalidate(store.Id, null);
                return Task.FromResult(HookResult.Ok);
            }
            else if (entry.Entity is Language lang && entry.InitialState != EntityState.Added)
            {
                InvalidateAll();
                return Task.FromResult(HookResult.Ok);
            }
            else if (entry.Entity is Setting setting)
            {
                if (setting.Name.EqualsNoCase(TypeHelper.NameOf<LocalizationSettings>(x => x.DefaultLanguageRedirectBehaviour)))
                {
                    InvalidateAll();
                }

                return Task.FromResult(HookResult.Ok);
            }

            return Task.FromResult(HookResult.Void);
        }

        #endregion

        public virtual async Task<XmlSitemapPartition> GetSitemapPartAsync(int index = 0)
        {
            return await GetSitemapPartAsync(index, false);
        }

        private async Task<XmlSitemapPartition> GetSitemapPartAsync(int index, bool isRetry)
        {
            Guard.NotNegative(index, nameof(index));

            var store = _services.StoreContext.CurrentStore;
            var language = _services.WorkContext.WorkingLanguage;

            var exists = TryGetSitemapFile(store.Id, language.Id, index, out var file);

            if (exists)
            {
                return new XmlSitemapPartition
                {
                    Index = index,
                    Name = file.Name,
                    LanguageId = language.Id,
                    StoreId = store.Id,
                    ModifiedOnUtc = file.LastModified.UtcDateTime,
                    Stream = await file.OpenReadAsync()
                };
            }

            if (isRetry)
            {
                var msg = "Could not generate XML sitemap. Index: {0}, Date: {1}".FormatInvariant(index, DateTime.UtcNow);
                Logger.Error(msg);
                throw new Exception(msg);
            }

            if (index > 0)
            {
                // File with index greater 0 has been requested, but it does not exist.
                // Now we have to determine whether just the passed index is out of range
                // or the files have never been created before.
                // If the main file (index 0) exists, the action should return NotFoundResult,
                // otherwise the rebuild process should be started or waited for.

                if (TryGetSitemapFile(store.Id, language.Id, 0, out file))
                {
                    throw new IndexOutOfRangeException("The sitemap file '{0}' does not exist.".FormatInvariant(file.Name));
                }
            }

            // The main sitemap document with index 0 does not exist, meaning: the whole sitemap
            // needs to be created and cached by partitions.

            var wasRebuilding = false;
            var lockFilePath = GetLockFilePath(store.Id, language.Id);

            while (await IsRebuildingAsync(lockFilePath))
            {
                // The rebuild process is already running, either started
                // by the task scheduler or another HTTP request.
                // We should wait for completion.

                wasRebuilding = true;
                await Task.Delay(1000);
            }

            if (!wasRebuilding)
            {
                // No lock. Rebuild now.
                var buildContext = new XmlSitemapBuildContext(store, new[] { language }, _services.SettingFactory, _services.StoreContext.IsSingleStoreMode())
                {
                    CancellationToken = _asyncRunner.AppShutdownCancellationToken
                };

                await RebuildAsync(buildContext);
            }

            // DRY: call self to get sitemap partition object
            return await GetSitemapPartAsync(index, true);
        }

        private bool TryGetSitemapFile(int storeId, int languageId, int index, out IFile file)
        {
            var path = BuildSitemapFilePath(storeId, languageId, index);

            file = _tenantRoot.GetFile(path);

            return file.Exists;
        }

        private string BuildSitemapFilePath(int storeId, int languageId, int index)
        {
            var fileName = SitemapFileNamePattern.FormatInvariant(index);
            return PathUtility.Join(BuildSitemapDirPath(storeId, languageId), fileName);
        }

        private string BuildSitemapDirPath(int? storeId, int? languageId)
        {
            if (storeId == null)
            {
                return _baseDir;
            }

            if (languageId == null)
            {
                return PathUtility.Join(_baseDir, storeId.ToStringInvariant());
            }
            else
            {
                return PathUtility.Join(_baseDir, storeId.ToStringInvariant(), languageId.ToStringInvariant());
            }
        }

        private string GetLockFilePath(int storeId, int languageId)
        {
            var fileName = LockFileNamePattern.FormatInvariant(storeId, languageId);
            return PathUtility.Join(_baseDir, fileName);
        }

        public virtual async Task RebuildAsync(XmlSitemapBuildContext ctx)
        {
            Guard.NotNull(ctx, nameof(ctx));

            var languageData = new Dictionary<int, LanguageData>();

            foreach (var language in ctx.Languages)
            {
                var lockFilePath = GetLockFilePath(ctx.Store.Id, language.Id);

                if (_lockFileManager.TryAcquireLock(lockFilePath, out var lockFile))
                {
                    // Process only languages that are unlocked right now
                    // It is possible that an HTTP request triggered the generation
                    // of a language specific sitemap.

                    try
                    {
                        var sitemapDir = BuildSitemapDirPath(ctx.Store.Id, language.Id);
                        var data = new LanguageData
                        {
                            Store = ctx.Store,
                            Language = language,
                            LockHandle = lockFile,
                            LockFilePath = lockFilePath,
                            TempDir = sitemapDir + "~",
                            FinalDir = sitemapDir,
                            BaseUrl = await BuildBaseUrlAsync(ctx.Store, language)
                        };

                        _tenantRoot.TryDeleteDirectory(data.TempDir);
                        _tenantRoot.TryCreateDirectory(data.TempDir);

                        languageData[language.Id] = data;
                    }
                    catch
                    {
                        await lockFile.ReleaseAsync();
                        throw;
                    }
                }
            }

            if (languageData.Count == 0)
            {
                Logger.Warn("XML sitemap rebuild already in process.");
                return;
            }

            var languages = languageData.Values.Select(x => x.Language);
            var languageIds = languages.Select(x => x.Id).Concat(new[] { 0 }).ToArray();

            // All sitemaps grouped by language
            var sitemaps = new Multimap<int, XmlSitemapNode>();

            var compositeFileLock = new AsyncActionDisposable(async () =>
            {
                foreach (var data in languageData.Values)
                {
                    await data.LockHandle.ReleaseAsync();
                }
            });

            await using (compositeFileLock)
            {
                // Impersonate
                var prevCustomer = _services.WorkContext.CurrentCustomer;
                // no need to vary xml sitemap by customer roles: it's relevant to crawlers only.
                _services.WorkContext.CurrentCustomer = (await _customerService.GetCustomerBySystemNameAsync(SystemCustomerNames.SearchEngine, false)) ?? prevCustomer;

                try
                {
                    var nodes = new List<XmlSitemapNode>();

                    var providers = CreateProviders(ctx);
                    var total = await providers.SelectAwait(x => x.GetTotalCountAsync()).SumAsync(ctx.CancellationToken);
                    var totalSegments = (int)Math.Ceiling(total / (double)MaximumSiteMapNodeCount);
                    var hasIndex = totalSegments > 1;
                    var indexNodes = new Multimap<int, XmlSitemapNode>();
                    var segment = 0;
                    var numProcessed = 0;

                    CheckSitemapCount(totalSegments);

                    using (new DbContextScope(_db, autoDetectChanges: false, forceNoTracking: true, lazyLoading: false))
                    {
                        var entities = EnlistEntitiesAsync(providers);

                        await foreach (var batch in entities.ChunkAsync(MaximumSiteMapNodeCount, ctx.CancellationToken))
                        {
                            if (ctx.CancellationToken.IsCancellationRequested)
                            {
                                break;
                            }

                            segment++;
                            numProcessed = segment * MaximumSiteMapNodeCount;
                            ctx.ProgressCallback?.Invoke(numProcessed, total, "{0} / {1}".FormatCurrent(numProcessed, total));

                            var slugs = await GetUrlRecordCollectionsForBatchAsync(batch.Select(x => x.Entry).ToList(), languageIds);

                            foreach (var data in languageData.Values)
                            {
                                var language = data.Language;
                                var baseUrl = data.BaseUrl;

                                // Create all node entries for this segment
                                var entries = batch
                                    .Where(x => x.Entry.LanguageId.GetValueOrDefault() == 0 || x.Entry.LanguageId.Value == language.Id)
                                    .Select(x => x.Provider.CreateNode(_linkGenerator, baseUrl, x.Entry, slugs[x.Entry.EntityName], language));
                                sitemaps[language.Id].AddRange(entries.Where(x => x != null));

                                // Create index node for this segment/language combination
                                if (hasIndex)
                                {
                                    indexNodes[language.Id].Add(new XmlSitemapNode
                                    {
                                        LastMod = sitemaps[language.Id].Select(x => x.LastMod).Where(x => x.HasValue).DefaultIfEmpty().Max(),
                                        Loc = GetSitemapIndexUrl(segment, baseUrl),
                                    });
                                }

                                if (segment % 5 == 0 || segment == totalSegments)
                                {
                                    // Commit every 5th segment (10.000 nodes) temporarily to disk to minimize RAM usage
                                    var documents = GetSiteMapDocuments((IReadOnlyCollection<XmlSitemapNode>)sitemaps[language.Id]);
                                    await SaveTempAsync(documents, data, segment - documents.Count + (hasIndex ? 1 : 0));

                                    documents.Clear();
                                    sitemaps.RemoveAll(language.Id);
                                }
                            }

                            slugs.Clear();
                        }

                        // Process custom nodes
                        if (!ctx.CancellationToken.IsCancellationRequested)
                        {
                            ctx.ProgressCallback?.Invoke(numProcessed, total, "Processing custom nodes".FormatCurrent(numProcessed, total));
                            await ProcessCustomNodesAsync(ctx, sitemaps);

                            foreach (var data in languageData.Values)
                            {
                                if (sitemaps.ContainsKey(data.Language.Id) && sitemaps[data.Language.Id].Count > 0)
                                {
                                    var documents = GetSiteMapDocuments((IReadOnlyCollection<XmlSitemapNode>)sitemaps[data.Language.Id]);
                                    await SaveTempAsync(documents, data, (segment + 1) - documents.Count + (hasIndex ? 1 : 0));
                                }
                                else if (segment == 0)
                                {
                                    // Ensure that at least one entry exists. Otherwise,
                                    // the system will try to rebuild again.
                                    var homeNode = new XmlSitemapNode { LastMod = DateTime.UtcNow, Loc = data.BaseUrl };
                                    var documents = GetSiteMapDocuments(new List<XmlSitemapNode> { homeNode });
                                    await SaveTempAsync(documents, data, 0);
                                }

                            }
                        }
                    }

                    ctx.CancellationToken.ThrowIfCancellationRequested();

                    ctx.ProgressCallback?.Invoke(totalSegments, totalSegments, "Finalizing...'");

                    foreach (var data in languageData.Values)
                    {
                        // Create index documents (if any)
                        if (hasIndex && indexNodes.Count > 0)
                        {
                            var indexDocument = CreateSitemapIndexDocument(indexNodes[data.Language.Id]);
                            await SaveTempAsync(new List<string> { indexDocument }, data, 0);
                        }

                        // Save finally (actually renames temp folder)
                        await SaveFinalAsync(data);
                    }
                }
                finally
                {
                    // Undo impersonation
                    _services.WorkContext.CurrentCustomer = prevCustomer;
                    sitemaps.Clear();

                    foreach (var data in languageData.Values)
                    {
                        if (_tenantRoot.DirectoryExists(data.TempDir))
                        {
                            _tenantRoot.TryDeleteDirectory(data.TempDir);
                        }
                    }
                }
            }
        }

        private async Task<string> BuildBaseUrlAsync(Store store, Language language)
        {
            var host = _services.StoreContext.CurrentStore.GetHost();

            var localizationSettings = await _services.SettingFactory.LoadSettingsAsync<LocalizationSettings>(store.Id);
            if (localizationSettings.SeoFriendlyUrlsForLanguagesEnabled)
            {
                var defaultLangId = await _languageService.GetMasterLanguageIdAsync(store.Id);
                if (language.Id != defaultLangId || localizationSettings.DefaultLanguageRedirectBehaviour < DefaultLanguageRedirectBehaviour.StripSeoCode)
                {
                    host += language.GetTwoLetterISOLanguageName() + '/';
                }
            }

            return host;
        }

        private async Task SaveTempAsync(List<string> documents, LanguageData data, int start)
        {
            for (int i = 0; i < documents.Count; i++)
            {
                // Save segment to disk
                var fileName = SitemapFileNamePattern.FormatInvariant(i + start);
                var filePath = PathUtility.Join(data.TempDir, fileName);

                await _tenantRoot.WriteAllTextAsync(filePath, documents[i]);
            }
        }

        private async Task SaveFinalAsync(LanguageData data)
        {
            // Delete current sitemap dir
            _tenantRoot.TryDeleteDirectory(data.FinalDir);

            // Move/Rename new (temp) dir to current
            _tenantRoot.MoveEntry(data.TempDir, data.FinalDir);

            int retries = 0;
            while (!TryGetSitemapFile(data.Store.Id, data.Language.Id, 0, out _))
            {
                if (retries > 20)
                {
                    break;
                }

                // IO breathe: directly after a folder rename a file check fails. Wait a sec...
                await Task.Delay(500);
                retries++;
            }
        }

        private static async IAsyncEnumerable<NodeEntry> EnlistEntitiesAsync(XmlSitemapProvider[] providers, [EnumeratorCancellation] CancellationToken cancelToken = default)
        {
            foreach (var provider in providers)
            {
                await foreach (var entry in provider.EnlistAsync(cancelToken))
                {
                    yield return new NodeEntry { Entry = entry, Provider = provider };
                }
            }
        }

        private async Task<Dictionary<string, UrlRecordCollection>> GetUrlRecordCollectionsForBatchAsync(IEnumerable<NamedEntity> batch, int[] languageIds)
        {
            var result = new Dictionary<string, UrlRecordCollection>();

            if (batch.First().EntityName == "Product")
            {
                // Nothing comes after product
                int min = batch.Last().Id;
                int max = batch.First().Id;

                result["Product"] = await _urlService.GetUrlRecordCollectionAsync("Product", languageIds, new[] { min, max }, true, true);
            }

            var entityGroups = batch.ToMultimap(x => x.EntityName, x => x.Id);
            foreach (var group in entityGroups)
            {
                var isRange = group.Key == "Product";
                var entityIds = isRange ? new[] { group.Value.Last(), group.Value.First() } : group.Value.ToArray();

                result[group.Key] = await _urlService.GetUrlRecordCollectionAsync(group.Key, languageIds, entityIds, isRange, isRange);
            }

            return result;
        }

        protected virtual List<string> GetSiteMapDocuments(IReadOnlyCollection<XmlSitemapNode> nodes)
        {
            int siteMapCount = (int)Math.Ceiling(nodes.Count / (double)MaximumSiteMapNodeCount);
            CheckSitemapCount(siteMapCount);

            var siteMaps = Enumerable
                .Range(0, siteMapCount)
                .Select(x =>
                {
                    return new KeyValuePair<int, IEnumerable<XmlSitemapNode>>(
                        x + 1,
                        nodes.Skip(x * MaximumSiteMapNodeCount).Take(MaximumSiteMapNodeCount));
                });

            var siteMapDocuments = new List<string>(siteMapCount);

            foreach (var kvp in siteMaps)
            {
                siteMapDocuments.Add(GetSitemapDocument(kvp.Value));
            }

            return siteMapDocuments;
        }

        /// <summary>
        /// Gets the sitemap XML document for the specified set of nodes.
        /// </summary>
        /// <param name="nodes">The sitemap nodes.</param>
        /// <returns>The sitemap XML document for the specified set of nodes.</returns>
        private string GetSitemapDocument(IEnumerable<XmlSitemapNode> nodes)
        {
            //var languages = _languageService.GetAllLanguages();

            XNamespace ns = SitemapsNamespace;
            XNamespace xhtml = XhtmlNamespace;

            XElement root = new XElement(
                ns + "urlset",
                new XAttribute(XNamespace.Xmlns + "xhtml", xhtml));

            foreach (var node in nodes)
            {
                // url
                var xel = new XElement
                (
                    ns + "url",
                    // url/loc
                    new XElement(ns + "loc", node.Loc),
                    // url/lastmod
                    node.LastMod == null ? null : new XElement(
                        ns + "lastmod",
                        node.LastMod.Value.ToLocalTime().ToString("yyyy-MM-ddTHH:mm:sszzz")),
                    // url/changefreq
                    node.ChangeFreq == null ? null : new XElement(
                        ns + "changefreq",
                        node.ChangeFreq.Value.ToString().ToLowerInvariant()),
                    // url/priority
                    node.Priority == null ? null : new XElement(
                        ns + "priority",
                        node.Priority.Value.ToString("F1", CultureInfo.InvariantCulture))
                );

                if (node.Links != null)
                {
                    foreach (var entry in node.Links)
                    {
                        // url/xhtml:link[culture]
                        xel.Add(new XElement
                        (
                            xhtml + "link",
                            new XAttribute("rel", "alternate"),
                            new XAttribute("hreflang", entry.Lang),
                            new XAttribute("href", entry.Href)
                        ));
                    }
                }

                root.Add(xel);
            }

            XDeclaration declaration = new XDeclaration("1.0", "UTF-8", "yes");
            XDocument document = new XDocument(root);
            var xml = declaration.ToString() + document.ToString(SaveOptions.DisableFormatting);
            CheckDocumentSize(xml);

            return xml;
        }

        /// <summary>
        /// Gets the sitemap index XML document, containing links to all the sitemap XML documents.
        /// </summary>
        /// <param name="nodes">The collection of sitemaps containing their index and nodes.</param>
        /// <returns>The sitemap index XML document, containing links to all the sitemap XML documents.</returns>
        private string CreateSitemapIndexDocument(IEnumerable<XmlSitemapNode> nodes)
        {
            XNamespace ns = SitemapsNamespace;

            XElement root = new XElement(ns + "sitemapindex");

            foreach (var node in nodes)
            {
                var xel = new XElement(
                    ns + "sitemap",
                    new XElement(ns + "loc", node.Loc),
                    node.LastMod.HasValue ?
                        new XElement(
                            ns + "lastmod",
                            node.LastMod.Value.ToLocalTime().ToString("yyyy-MM-ddTHH:mm:sszzz")) :
                        null);

                root.Add(xel);
            }

            var document = new XDocument(root);
            var xml = document.ToString(SaveOptions.DisableFormatting);
            CheckDocumentSize(xml);

            return xml;
        }

        private static string GetSitemapIndexUrl(int index, string baseUrl)
        {
            return baseUrl + SitemapIndexPathPattern.FormatInvariant(index);
        }

        private XmlSitemapProvider[] CreateProviders(XmlSitemapBuildContext context)
        {
            return _publishers
                .Select(x => x.Value.PublishXmlSitemap(context))
                .Where(x => x != null)
                .OrderBy(x => x.Order)
                .ToArray();
        }

        protected virtual Task ProcessCustomNodesAsync(XmlSitemapBuildContext ctx, Multimap<int, XmlSitemapNode> sitemaps)
        {
            // For inheritors
            return Task.CompletedTask;
        }

        /// <summary>
        /// Checks the size of the XML sitemap document. If it is over 10MB, logs an error.
        /// </summary>
        /// <param name="siteMapXml">The sitemap XML document.</param>
        private void CheckDocumentSize(string siteMapXml)
        {
            if (siteMapXml.Length >= MaximumSiteMapSizeInBytes)
            {
                Logger.Error(new InvalidOperationException($"Sitemap exceeds the maximum size of 10MB. This is because you have unusually long URL's. Consider reducing the MaximumSitemapNodeCount. Size:<{siteMapXml.Length}>"));
            }
        }

        /// <summary>
        /// Checks the count of the number of sitemaps. If it is over 50,000, logs an error.
        /// </summary>
        /// <param name="sitemapCount">The sitemap count.</param>
        private void CheckSitemapCount(int sitemapCount)
        {
            if (sitemapCount > MaximumSiteMapCount)
            {
                var ex = new InvalidOperationException($"Sitemap index file exceeds the maximum number of allowed sitemaps of 50,000. Count:<{sitemapCount}>");
                Logger.Warn(ex, ex.Message);
            }
        }

        public Task<bool> IsRebuildingAsync(int storeId, int languageId)
        {
            return IsRebuildingAsync(GetLockFilePath(storeId, languageId));
        }

        private Task<bool> IsRebuildingAsync(string lockFilePath)
        {
            return _lockFileManager.IsLockedAsync(lockFilePath);
        }

        public virtual bool IsGenerated(int storeId, int languageId)
        {
            return TryGetSitemapFile(storeId, languageId, 0, out _);
        }

        public virtual void Invalidate(int storeId, int? languageId)
        {
            var dir = BuildSitemapDirPath(storeId, languageId);
            _tenantRoot.TryDeleteDirectory(dir);
        }

        public virtual void InvalidateAll()
        {
            var dir = BuildSitemapDirPath(null, null);
            foreach (var subDir in _tenantRoot.EnumerateDirectories(dir).ToArray())
            {
                // Delete only directories, no lock files.
                subDir.Delete();
            }
        }

        #region Nested classes

        struct NodeEntry
        {
            public NamedEntity Entry { get; init; }
            public XmlSitemapProvider Provider { get; init; }
        }

        class LanguageData
        {
            public Store Store { get; init; }
            public Language Language { get; init; }
            public ILockHandle LockHandle { get; init; }
            public string LockFilePath { get; init; }
            public string TempDir { get; init; }
            public string FinalDir { get; init; }
            public string BaseUrl { get; init; }
        }

        #endregion
    }
}
